import asyncio
import random
from typing import Dict, List, Optional, Tuple, Union

import openai

from ai.database import AgentSchema, Credentials, MessageSchema, UsageSchema


class _Conversation:
    PRICE_PER_TOKEN = 0.002 / 1000

    def __init__(self, who: Optional[AgentSchema] = None) -> None:
        conversations = []
        if who is not None:
            conversations.append(who.instructions.dict())
        self._conversation: List[Dict] = conversations
        self.total_tokens = 0

    async def ask(self, prompt: str) -> str:
        msg = MessageSchema(role="user", content=prompt)
        self._conversation.append(msg.dict())
        msg = await self._send()
        return msg.content

    @property
    def total_cost(self) -> float:
        return self.total_tokens * self.PRICE_PER_TOKEN

    async def _send(self) -> MessageSchema:
        raise NotImplementedError

    def start(self, credentials: Credentials):
        raise NotImplementedError


class EchoConversation(_Conversation):
    async def _send(self) -> MessageSchema:
        num = random.random()
        await asyncio.sleep(num)
        msg = MessageSchema(role="assistant", content=f"mock response {num}")
        self._conversation.append(msg.dict())
        self.total_tokens += 0
        template = " - {role}: {content}"
        text_convo = "\n".join(template.format(**m) for m in self._conversation)
        return MessageSchema(role="assistant", content=text_convo)

    def start(self, credentials: Credentials):
        pass


class GPTConversation(_Conversation):
    MODEL = "gpt-3.5-turbo"

    async def _send(self) -> MessageSchema:
        res = await openai.ChatCompletion.acreate(
            model=self.MODEL, messages=self._conversation
        )
        msg, usage = parse_response(res)
        self._conversation.append(msg.dict())
        self.total_tokens += usage.total_tokens
        return msg

    def start(self, credentials: Credentials):
        openai.organization = credentials.id_
        openai.api_key = credentials.api_key


Conversation = Union[EchoConversation, GPTConversation]


def parse_response(res: Dict) -> Tuple[MessageSchema, UsageSchema]:
    msg = MessageSchema(**res["choices"][0]["message"])
    usage = UsageSchema(**res["usage"])
    return msg, usage



import argparse
import sys
import tempfile
from pathlib import Path
from typing import List, Optional

from rich import print
from rich.prompt import Confirm, Prompt

from ai import ai, crud, database
from ai.database import session
from ai.tui import Chat

ROOT = Path.home() / ".chatgpt_tui"
DEFAULT_DB = ROOT / "db.json"

INITIAL_MENU = (
    "Speak with the default agent",
    "Select an agent from your saved list",
    "Create a new agent",
)


class Menu:
    def __init__(self, options: List, start_idx: int = 0):
        self.options = options
        self.start_idx = start_idx

    def get(self, i: int):
        return self.options[i - self.start_idx]

    @property
    def choices(self) -> List[str]:
        return [str(i) for i, _ in enumerate(self.options, start=self.start_idx)]

    def __rich__(self) -> str:
        display = [
            f"[bold magenta]{i}[/] - {option}"
            for i, option in enumerate(self.options, start=self.start_idx)
        ]
        return "\n".join(display)


def new_agent() -> database.AgentSchema:
    name = Prompt.ask("Enter a name for your agent")
    instructions = Prompt.ask("Enter instructions for your agent")
    agent = crud.new_agent(name, instructions)
    return agent


def first_time_setup():
    ok_to_setup = Confirm.ask(f"Adding folders to: {ROOT}.\nIs this okay?")
    if not ok_to_setup:
        raise SystemExit("Sorry about that")
    ROOT.mkdir()
    id_ = Prompt.ask("Enter your organization id")
    api_key = Prompt.ask("Enter your api-key")
    metadata = database.MetaDataSchema.latest(id_=id_, api_key=api_key)
    session().setup(metadata)


def start_debug_session():
    with tempfile.NamedTemporaryFile(suffix=".json") as tmp_db:
        db_path = Path(tmp_db.name)

        session.use_database(db_path)
        metadata = database.MetaDataSchema.latest(id_=1, api_key=1)
        session().setup(metadata)

        agent = get_agent()
        convo = ai.EchoConversation(agent)
        convo.start(credentials=None)

        app = Chat(conversation=convo)
        app.run()


def start_chat(agent: database.AgentSchema):
    convo = ai.GPTConversation(who=agent)
    credentials = crud.get_credentials()
    convo.start(credentials)
    app = Chat(conversation=convo)
    app.run()


def init_session(db_path):
    session.use_database(db_path)
    if not session().is_setup:
        first_time_setup()


def get_agent() -> Optional[database.AgentSchema]:
    agents = {agent.name: agent for agent in crud.all_agents()}
    options = ["Create New Agent", "Default Agent", *agents.keys()]
    menu = Menu(options=options)
    print(menu)
    choice = Prompt.ask("Select an agent", choices=menu.choices)
    assert choice.isnumeric(), "Choice should have been an int, something went wrong"
    choice = int(choice)
    if choice == 0:
        agent = new_agent()
    elif choice == 1:
        agent = None
    else:
        agent_name = menu.get(choice)
        agent = agents.get(agent_name)

    return agent


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Enter into a debug mode that doesn't send info to openai",
    )
    parser.add_argument(
        "--db",
        type=Path,
        default=DEFAULT_DB,
        help="The path to a database you want to use if not the default",
    )
    args = parser.parse_args()

    if args.debug:
        start_debug_session()
        return

    init_session(args.db)
    agent = get_agent()
    start_chat(agent)


if __name__ == "__main__":
    sys.exit(main())



from ai import database
from ai.database import session


def new_agent(name, instructions):
    with session() as db:
        agent = database.add_agent(name=name, instructions=instructions, db=db)
        db.commit()
    return agent


def all_agents():
    with session() as db:
        agents = db.agents.copy()
    return agents


def get_credentials() -> database.Credentials:
    with session() as db:
        credentials = database.Credentials(id_=db.id_, api_key=db.api_key)
        return credentials



import json
from functools import wraps
from io import TextIOWrapper
from pathlib import Path
from typing import Callable, List, Literal, Optional

from pydantic import BaseModel, validator

LATEST_VERSION = 1


class MessageSchema(BaseModel):
    role: Literal["user", "assistant", "system"]
    content: str


class UsageSchema(BaseModel):
    completion_tokens: int
    prompt_tokens: int
    total_tokens: int


class AgentSchema(BaseModel):
    name: str
    instructions: MessageSchema

    @validator("instructions")
    def role_must_be_system(cls, v):
        if v.role != "system":
            raise ValueError("Agent's role must be `system`")
        return v


class Credentials(BaseModel):
    id_: str
    api_key: str


class MetaDataSchema(Credentials):
    version: int

    @classmethod
    def latest(cls, **kwargs) -> "MetaDataSchema":
        return cls(version=LATEST_VERSION, **kwargs)


class DBSchema(MetaDataSchema):
    agents: List[AgentSchema]


def assert_connected(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(self: "Connection", *args, **kwargs):
        assert self.db is not None
        return func(self, *args, **kwargs)

    return wrapper


class Connection:
    def __init__(self) -> None:
        self._commit: bool = False
        self.schema = DBSchema
        self.db: Optional[DBSchema] = None

    def initalize(self, metadata: MetaDataSchema):
        self.db = self.schema(**metadata.dict(), agents=[])
        self.commit()

    def connect(self, handler: TextIOWrapper):
        _db = json.load(handler)
        metadata = MetaDataSchema(**_db)
        # Only support latest version for now
        assert metadata.version == LATEST_VERSION
        self.db = self.schema(**_db)

    @assert_connected
    def commit(self):
        self._commit = True

    @assert_connected
    def write(self, handler: TextIOWrapper):
        if self._commit:
            json.dump(self.db.dict(), handler, indent=2)

    @assert_connected
    def add(self, row: MessageSchema):
        self.agents.append(row)

    @assert_connected
    def __getattr__(self, attr: str):
        return getattr(self.db, attr)


class session:
    _db_path: Optional[Path] = None

    def __init__(self) -> None:
        self._connection: Optional[Connection] = None

    @classmethod
    def use_database(cls, db_path: Path):
        assert cls._db_path is None, f"Database {cls._db_path} is already being used"
        cls._db_path = db_path
        return cls()

    @property
    def db_path(self):
        assert self._db_path is not None, "A database file needs to be connected"
        return self._db_path

    @property
    def is_setup(self) -> bool:
        if not self.db_path.exists():
            return False

        with self as db:
            has_credentials = None not in [db.api_key, db.id_]

        return has_credentials

    def setup(self, metadata: MetaDataSchema):
        conn = Connection()
        conn.initalize(metadata)
        with open(self.db_path, "w") as f:
            conn.write(f)

    def _connect(self) -> Connection:
        assert self._connection is None, "Already connected"
        conn = Connection()
        conn.connect(self._handler)
        return conn

    def __enter__(self) -> Connection:
        self._handler = open(self.db_path, "r+")
        self._connection = self._connect()
        return self._connection

    def __exit__(self, exc_type, exc_value, exc_traceback):
        assert self._connection is not None, "No connection to exit"
        self._handler.seek(0)
        self._connection.write(self._handler)
        self._handler.close()
        self._connection = None

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            with self as db:
                res = func(*args, **kwargs, db=db)
                return res

        return wrapper


def add_agent(name: str, instructions: str, db: Connection):
    agent = AgentSchema(
        name=name, instructions=MessageSchema(role="system", content=instructions)
    )
    db.add(agent)
    return agent



import pathlib

from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / 'README.md').read_text()

# This call to setup() does all the work
setup(
    name="aitui",
    version="0.0.1",
    description="",
    long_description=README,
    long_description_content_type="text/markdown",
    url="",
    author="Ben Hammel",
    author_email="bdhammel@gmail.com",
    license="MIT",
    install_requires=(HERE / "requirements.txt").read_text().splitlines(),
    entry_points={
        'console_scripts': ['ai=ai.cli:main'],
    },
    packages=find_packages(),
    include_package_data=True
)



import subprocess
import tempfile
from dataclasses import dataclass
from enum import Enum

from rich.console import Console, ConsoleOptions, RenderResult
from rich.markdown import Markdown
from rich.padding import Padding
from rich.text import Text
from textual.app import App, ComposeResult
from textual.widgets import Footer, Header, Input, TextLog

VIM_CMD = ["vim", "+set backupcopy=yes", "+normal G$" "+startinsert"]


class VimMode(Enum):
    INS = "insert"
    CMD = "command"


def chain_action(*args):
    def _chain():
        for action in args:
            action()

    return _chain


@dataclass
class TextBlock:
    speaker: str
    message: str

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Text(f"{self.speaker}", style="magenta")
        yield Markdown(self.message)


class Prompt(Input):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mode = VimMode.INS
        self.background_mode = {
            VimMode.INS: self.styles.background,
            VimMode.CMD: "blue",
        }

    def on_input_submitted(self, message: Input.Submitted) -> None:
        message.value = message.value
        self.value = ""

    def action_cmd_mode(self) -> None:
        self.mode = VimMode.CMD
        self.styles.background = self.background_mode[self.mode]

    def action_ins_mode(self) -> None:
        self.mode = VimMode.INS
        self.styles.background = self.background_mode[self.mode]

    async def on_key(self, event) -> None:
        if event.key == "escape":
            event.prevent_default()
            event.stop()
            if self.mode is VimMode.INS:
                self.action_cmd_mode()
        elif self.mode is VimMode.CMD:
            event.prevent_default()
            event.stop()
            action = {
                "i": self.action_ins_mode,
                "a": chain_action(self.action_cursor_right, self.action_ins_mode),
                "A": chain_action(self.action_end, self.action_ins_mode),
                "h": self.action_cursor_left,
                "l": self.action_cursor_right,
                "w": self.action_cursor_right_word,
                "W": self.action_cursor_right_word,
                "b": self.action_cursor_left_word,
                "B": self.action_cursor_left_word,
                "x": self.action_delete_right,
                "X": self.action_delete_left,
                "D": self.action_delete_right_all,
                "$": self.action_end,
                "_": self.action_home,
                "0": self.action_home,
                "v": chain_action(self.action_vim, self.action_ins_mode),
            }.get(event.character, None)

            if action is not None:
                action()

    def action_vim(self) -> None:
        app = self.app
        assert app._driver is not None
        app._driver.stop_application_mode()
        initial_text = self.value
        try:
            with tempfile.NamedTemporaryFile(suffix=".txt") as tf:
                tf.write(bytes(initial_text, "UTF-8"))
                tf.flush()
                tf.seek(0, 2)
                # Need to create a separate backup copy
                # If we don't the edited text will not be saved into the current file
                subprocess.call([*VIM_CMD, tf.name])
                tf.seek(0)
                prompt = tf.read().decode("utf-8")
        finally:
            app.refresh()
            app._driver.start_application_mode()

        self.post_message_no_wait(Input.Submitted(sender=self, value=prompt))


class ConversationScreen(TextLog):
    def __init__(self, *args, **kwargs):
        super().__init__(wrap=True, highlight=True, markup=True, *args, **kwargs)

    async def append(self, message: TextBlock):
        self.write(Padding(message, (1, 1)))
        self.scroll_end(animate=True, duration=1)

    async def user_says(self, message: str):
        await self.append(TextBlock("User", message))

    async def agent_says(self, message: str):
        await self.append(TextBlock("Agent", message))


class Chat(App):
    CSS_PATH = "css/style.css"

    BINDINGS = [
        ("crtl+q", "quit", "close"),
    ]

    def __init__(self, conversation, *args, **kwargs):
        self.conversation = conversation
        super().__init__(*args, **kwargs)

    def compose(self) -> ComposeResult:
        yield Header()
        yield ConversationScreen(id="chat")
        yield Prompt(placeholder="Prompt", classes="prompt", id="prompt")
        yield Footer()

    def on_mount(self) -> None:
        """Called when app starts."""
        # Give the input focus, so we can start typing straight away
        self.query_one(Input).focus()

    async def on_input_submitted(self, message: Input.Submitted) -> None:
        prompt = message.value
        await self.ask(prompt)

    async def ask(self, prompt: str) -> None:
        if not prompt:
            return
        convo = self.query_one("#chat", ConversationScreen)
        await convo.user_says(prompt)
        reply = await self.conversation.ask(prompt)
        await convo.agent_says(reply)



import random

import pytest

from ai.database import (
    LATEST_VERSION,
    Connection,
    DBSchema,
    MetaDataSchema,
    add_agent,
    session,
)


def test_initial_setup(empty_db_file):
    db_file = empty_db_file
    sess = session.use_database(db_file)
    assert not sess.is_setup

    id_ = random.random()
    api_key = random.random()
    expected = DBSchema(version=LATEST_VERSION, id_=id_, api_key=api_key, agents=[])

    metadata = MetaDataSchema.latest(id_=id_, api_key=api_key)
    sess.setup(metadata)

    assert sess.is_setup

    actual = DBSchema.parse_file(db_file)
    assert actual.dict() == expected.dict()


@pytest.mark.usefixtures("active_session")
def test_context_handler():
    sess = session()
    db_path = sess.db_path
    assert sess.is_setup
    num_agents = 5

    with sess as db:
        for i in range(num_agents):
            add_agent(name=f"agent-{i}", instructions=i, db=db)
        expected = db.db.copy()
        db.commit()

    actual = DBSchema.parse_file(db_path)
    assert len(actual.agents) == num_agents
    assert actual.dict() == expected.dict()


@pytest.mark.usefixtures("active_session")
def test_data_not_saved_if_no_commit():
    sess = session()
    db_path = sess.db_path
    assert sess.is_setup

    with sess as db:
        add_agent(name="agent", instructions="dne", db=db)

    actual = DBSchema.parse_file(db_path)
    assert len(actual.agents) == 0


@pytest.mark.usefixtures("active_session")
def test_decorator():
    sess = session()
    db_path = sess.db_path
    num_agents = 5

    @sess
    def foo(bar, *, db):
        add_agent(name="agent", instructions=bar, db=db)
        expected = db.db.copy()
        db.commit()
        return expected

    for i in range(num_agents):
        expected = foo(i)

    actual = DBSchema.parse_file(db_path)
    assert len(actual.agents) == num_agents
    assert actual.dict() == expected.dict()


def test_assert_connection():
    conn = Connection()

    with pytest.raises(AssertionError):
        conn.commit()

    with pytest.raises(AssertionError):
        conn.add("foo")

    with pytest.raises(AssertionError):
        conn.dne
